package at.ac.uibk.gitsearch.service;

import java.io.InputStream;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.codeability.sharing.plugins.api.ShoppingBasket;
import org.codeability.sharing.plugins.api.ShoppingBasket.ExerciseInfo;
import org.codeability.sharing.plugins.api.ShoppingBasket.UserInfo;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider.GitLabAccessInfo;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.SearchResultDTO;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class ShoppingBasketService {
	
	@Autowired
	private PluginManagementService pluginManagementService;
	@Autowired
	private GitLabRepository gitLabRepository;
	
	public static class ShoppingBasketRedirectInfoDTO {
		private String redirectURL;
		/**
		 * @return the redirectURL
		 */
		public String getRedirectURL() {
			return redirectURL;
		}
		/**
		 * @param redirectURL the redirectURL to set
		 */
		public void setRedirectURL(String redirectURL) {
			this.redirectURL = redirectURL;
		}
	}
	
	public static class ShoppingBasketInfoDTO {
		private SearchResultDTO[] itemInfos;
		private String plugin;
		private String action;
		
		public ShoppingBasketInfoDTO(SearchResultDTO[] itemInfos) {
			super();
			this.itemInfos = itemInfos;
		}
		
		public ShoppingBasketInfoDTO() {
			// JSON
		}
		
		

		/**
		 * @return the itemInfos
		 */
		public SearchResultDTO[] getItemInfos() {
			return itemInfos;
		}

		/**
		 * @param itemInfos the itemInfos to set
		 */
		public void setItemInfos(SearchResultDTO[] itemInfos) {
			this.itemInfos = itemInfos;
		}



		/**
		 * @return the plugin
		 */
		public String getPlugin() {
			return plugin;
		}

		/**
		 * @param plugin the plugin to set
		 */
		public void setPlugin(String plugin) {
			this.plugin = plugin;
		}

		/**
		 * @return the action
		 */
		public String getAction() {
			return action;
		}

		/**
		 * @param action the action to set
		 */
		public void setAction(String action) {
			this.action = action;
		}



		public static class ItemInfoDTO {
		String exerciseID;
		String gitLabURL;
		
		public ItemInfoDTO(String exerciseID, String gitLabURl) {
			super();
			this.exerciseID = exerciseID;
			this.gitLabURL = gitLabURl;
		}
		
		public ItemInfoDTO() {
			// JSON
		}
		/**
		 * @return the exerciseIDs
		 */
		public String getExerciseID() {
			return exerciseID;
		}
		/**
		 * @param exerciseIDs the exerciseIDs to set
		 */
		public void setExerciseID(String exerciseID) {
			this.exerciseID = exerciseID;
		}
		/**
		 * @return the gitLabURls
		 */
		public String getGitLabURL() {
			return gitLabURL;
		}
		/**
		 * @param gitLabURL the gitLabURls to set
		 */
		public void setGitLabURL(String gitLabURL) {
			this.gitLabURL = gitLabURL;
		}
		}
	}
	
	/**
	 * holds further info for shopping basket.
	 * @author Michael Breu
	 *
	 */
	public static class ExtendedShoppingBasket {
		public ExtendedShoppingBasket(ShoppingBasketInfoDTO shoppingBasket, Optional<GitLabAccessInfo> gitLabAccessInfo) {
			super();
			this.shoppingBasket = shoppingBasket;
			this.gitLabAccessInfo = gitLabAccessInfo;
		}
		public ExtendedShoppingBasket() {
		}
		
		private ShoppingBasketInfoDTO shoppingBasket;
		private Optional<GitLabAccessInfo> gitLabAccessInfo;
		
		/**
		 * @return the shoppingBasket
		 */
		public ShoppingBasketInfoDTO getShoppingBasket() {
			return shoppingBasket;
		}
		/**
		 * @param shoppingBasket the shoppingBasket to set
		 */
		public void setShoppingBasket(ShoppingBasketInfoDTO shoppingBasket) {
			this.shoppingBasket = shoppingBasket;
		}
		/**
		 * @return the gitLabAccessInfo
		 */
		public Optional<GitLabAccessInfo> getGitLabAccessInfo() {
			return gitLabAccessInfo;
		}
		/**
		 * @param gitLabAccessInfo the gitLabAccessInfo to set
		 */
		public void setGitLabAccessInfo(Optional<GitLabAccessInfo> gitLabAccessInfo) {
			this.gitLabAccessInfo = gitLabAccessInfo;
		}
	}
	

	private final Logger log = LoggerFactory.getLogger(ShoppingBasketService.class);


	public ShoppingBasketService() {

	}
	
	public ShoppingBasket getBasket(String basketToken) {
		if(basketToken==null)  {
			log.warn("Basket for null not found" );
			return null;
		}
		final @Nullable ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
		if(basketInfo==null) {
			log.warn("Basket {} not found", basketToken );
			return null;
		}
		
		List<ExerciseInfo> result = new ArrayList<>();
		for(SearchResultDTO hit: basketInfo.getShoppingBasket().getItemInfos()) {
			result.add(convertFrom(hit));
		}
		UserInfo userInfo = null;
		final Optional<GitLabAccessInfo> gitLabAccessInfo = basketInfo.getGitLabAccessInfo();
		if(gitLabAccessInfo.isPresent() && gitLabAccessInfo.get().getUser().isPresent()) {
			userInfo = new UserInfo(gitLabAccessInfo.get().getUser().get().getUsername());
		}
		final ShoppingBasket shoppingBasket = new ShoppingBasket(userInfo,  result.toArray(new ExerciseInfo[] {}));
		log.info("Basket {} is delivered", basketToken );
		return shoppingBasket;
	}
	
	private ExerciseInfo convertFrom(SearchResultDTO hit) {
		ExerciseInfo ex = new ExerciseInfo();
		ex.gitLabProjectId = hit.getProject().getProject_id();
		ex.gitLabURI = hit.getRepositoryURL();
		ex.keywords = Arrays.asList(hit.getKeyword());
		ex.title = hit.getTitle();
		return ex;
	}
	
	public InputStream getRepositoryZip(String basketToken, int position) throws GitLabApiException {
		final @Nullable ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
		ShoppingBasket basket = getBasket( basketToken);
        final GitLabApi gitLabApi = basketInfo==null?gitLabRepository.getGitLabApi(Optional.empty()):gitLabRepository.getGitLabApi(basketInfo.getGitLabAccessInfo());
		return gitLabApi.getRepositoryApi().getRepositoryArchive(basket.exerciseInfo[position].gitLabProjectId, "HEAD", "zip");
	}
	
	public final static Duration BASKET_EXPIRY_INTERVAL = Duration.ofMinutes(10);
	/**
	 * this is a cache for shopping baskets. Entries expires after BASKET_EXPIRY_INTERVAL.
	 * TODO: Store in Database, to allow multiple application instances.
	 */
	protected LoadingCache<String, ExtendedShoppingBasket> basketCache = CacheBuilder.newBuilder()
		    .maximumSize(10000)
		    .expireAfterWrite(BASKET_EXPIRY_INTERVAL)
		    .build(
		        new CacheLoader<String, ExtendedShoppingBasket>() {
		          public ExtendedShoppingBasket load(String key)  {
		        	  return null;
		          }
		        });
	
	public ShoppingBasketRedirectInfoDTO getRedirectInfo(ShoppingBasketInfoDTO basket, String baseURL) {
		
		UUID random = UUID.randomUUID();
		basketCache.put(random.toString(), new ExtendedShoppingBasket(basket, gitLabRepository.getGitLabAccessInfo()));
		
		ShoppingBasketRedirectInfoDTO result = new ShoppingBasketRedirectInfoDTO();
		
		result.setRedirectURL(
				pluginManagementService.getRedirectURL(basket.getPlugin(), basket.getAction())
				+ "/"+random.toString()
				+ "?returnURL=" + baseURL 
				+ "&apiBaseURL=" + baseURL+"/api/pluginIF/v0.1");
		
		return result;
		
	}

}
